@@ -10,24 +10,28 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!IMPL")
+(in-package "SB-IMPL")
 
 ;;; various environment inquiries
 
-(!defvar *features*
-   ;; GCC_TLS is not a Lisp feature- it's just freeloading off the means
-   ;; by which additional #defines get into "genesis/config.h".
-   ;; Literally nothing except C code tests for it.
-   '#.(remove-if (lambda (x)
-                   (member x '(:gcc-tls)))
-                 sb!xc:*features*)
+;;; Compute the target's value of CL:*FEATURES* be deleting non-target features.
+;;; - :CONS-PROFILING   is used only to assign a compiler policy which persists
+;;;                     into the default baseline policy. It has no relevance post-build
+;;;                     in as much as policy can be changed later arbitrarily.
+;;; - :GCC-TLS          is not a Lisp feature- it's just freeloading off the means
+;;;                     by which additional #defines get into "genesis/config.h".
+;;; - :SB-AFTER-XC-CORE is merely a build option.
+;;; - :SB-XC            indicates the build phase and must not persist.
+(defparameter *features*
+   '#.(remove-if (lambda (x) (member x '(:cons-profiling :gcc-tls :sb-after-xc-core :sb-xc)))
+                 sb-xc:*features*)
   "a list of symbols that describe features provided by the
    implementation")
 
 (defun machine-instance ()
   "Return a string giving the name of the local machine."
-  #!+win32 (sb!win32::get-computer-name)
-  #!-win32 (truly-the simple-string (sb!unix:unix-gethostname)))
+  #+win32 (sb-win32::get-computer-name)
+  #-win32 (truly-the simple-string (sb-unix:unix-gethostname)))
 
 (declaim (type (or null string) *machine-version*))
 (defvar *machine-version*)
@@ -135,8 +139,9 @@ the file system."
 
 ;;; DEFmumble helpers
 
-(defun %defglobal (name value boundp doc docp source-location)
-  (%compiler-defglobal name :always-bound value (not boundp))
+(defun %defglobal (name value source-location &optional (doc nil docp))
+  (%compiler-defglobal name :always-bound
+                       (not (unbound-marker-p value)) value)
   (when docp
     (setf (documentation name 'variable) doc))
   (when source-location
@@ -163,23 +168,28 @@ the file system."
     (setf (info :source-location :variable var) source-location))
   var)
 
-(defun %defun (name def &optional inline-lambda dxable-args)
+(defun %defun (name def &optional inline-lambda extra-info)
   (declare (type function def))
   ;; should've been checked by DEFMACRO DEFUN
   (aver (legal-fun-name-p name))
-  (sb!c:%compiler-defun name nil inline-lambda dxable-args)
-  (when (fboundp name)
-    (warn 'redefinition-with-defun :name name :new-function def))
+  ;; If a warning handler decides to disallow this redefinition
+  ;; by nonlocally exiting, then we'll skip the rest of this stuff.
+  (when (and (fboundp name)
+             *type-system-initialized*)
+    (handler-bind (((satisfies sb-c::handle-condition-p)
+                     #'sb-c::handle-condition-handler))
+      (warn 'redefinition-with-defun :name name :new-function def)))
+  (sb-c:%compiler-defun name nil inline-lambda extra-info)
   (setf (fdefinition name) def)
   ;; %COMPILER-DEFUN doesn't do this except at compile-time, when it
   ;; also checks package locks. By doing this here we let (SETF
   ;; FDEFINITION) do the load-time package lock checking before
   ;; we frob any existing inline expansions.
-  (sb!c::%set-inline-expansion name nil inline-lambda dxable-args)
-  (sb!c::note-name-defined name :function)
+  (sb-c::%set-inline-expansion name nil inline-lambda extra-info)
+  (sb-c::note-name-defined name :function)
   name)
 
-(in-package "SB!C")
+(in-package "SB-C")
 
 (defun real-function-name (name)
   ;; Resolve the actual name of the function named by NAME
@@ -247,8 +257,8 @@ version 1[.0.0...] or greater."
               (lisp-implementation-version)
               subversions))))
 
-(!defvar sb!pcl::*!docstrings* nil)
+(defparameter sb-pcl::*!docstrings* nil)
 (defun (setf documentation) (string name doc-type)
   (declare (type (or null string) string))
-  (push (list string name doc-type) sb!pcl::*!docstrings*)
+  (push (list string name doc-type) sb-pcl::*!docstrings*)
   string)
