@@ -46,7 +46,7 @@
 ;;; fail to infer (< I FIXNUM): it does not understand that this
 ;;; constraint follows from (TYPEP I (INTEGER 0 0)).
 
-(in-package "SB!C")
+(in-package "SB-C")
 
 ;;; *CONSTRAINT-UNIVERSE* gets bound in IR1-PHASES to a fresh,
 ;;; zero-length, non-zero-total-size vector-with-fill-pointer.
@@ -65,7 +65,7 @@
   ;;     X is a LAMBDA-VAR and Y is a CTYPE. The value of X is
   ;;     constrained to be of type Y.
   ;;
-  ;; > or <
+  ;; >, <, or =
   ;;     X is a lambda-var and Y is a CTYPE. The relation holds
   ;;     between X and some object of type Y.
   ;;
@@ -78,14 +78,24 @@
   ;;     (array-in-bounds-p array 10) X can be either the lambda-var
   ;;     of ARRAY or VAR, while Y is either the lambda-var of VAR or a
   ;;     constant.
-  (kind nil :type (member typep < > eql
-                          array-in-bounds-p))
+  (kind nil :type (member typep < > = eql
+                          array-in-bounds-p
+                          equality))
   ;; The operands to the relation.
   (x nil :type lambda-var)
   (y nil :type constraint-y)
   ;; If true, negates the sense of the constraint, so the relation
   ;; does *not* hold.
   (not-p nil :type boolean))
+
+(defstruct (equality-constraint
+            (:include constraint)
+            (:constructor make-equality-constraint
+                (number operator x y not-p
+                 &aux (kind 'equality))))
+
+  (operator nil :type symbol))
+
 
 ;;; Historically, CMUCL and SBCL have used a sparse set implementation
 ;;; for which most operations are O(n) (see sset.lisp), but at the
@@ -163,9 +173,6 @@
     (or (= (conset-min conset) (conset-max conset))
         (not (find 1 (conset-vector conset)
                    :start (conset-min conset)
-                   ;; the :end argument can be commented out when
-                   ;; bootstrapping on a < 1.0.9 SBCL errors out with
-                   ;; a full call to DATA-VECTOR-REF-WITH-OFFSET.
                    :end (conset-max conset)))))
 
   (defun copy-conset (conset)
@@ -300,7 +307,7 @@
   (etypecase y
     (ctype
        (awhen (lambda-var-ctype-constraints x)
-         (dolist (con (gethash (sb!kernel::type-class-info y) it) nil)
+         (dolist (con (gethash (sb-kernel::type-class-info y) it) nil)
            (when (and (eq (constraint-kind con) kind)
                       (eq (constraint-not-p con) not-p)
                       (type= (constraint-y con) y))
@@ -345,7 +352,7 @@
       (ctype
        (let ((index (ensure-hash (lambda-var-ctype-constraints x)))
              (vec   (ensure-vec  (lambda-var-inheritable-constraints x))))
-         (push con (gethash (sb!kernel::type-class-info y) index))
+         (push con (gethash (sb-kernel::type-class-info y) index))
          (vector-push-extend con vec)))
       (lvar
        (let ((index (ensure-hash (lambda-var-eq-constraints x))))
@@ -545,7 +552,7 @@
 
 ;;; If REF is to a LAMBDA-VAR with CONSTRAINTs (i.e. we can do flow
 ;;; analysis on it), then return the LAMBDA-VAR, otherwise NIL.
-#!-sb-fluid (declaim (inline ok-ref-lambda-var))
+#-sb-fluid (declaim (inline ok-ref-lambda-var))
 (defun ok-ref-lambda-var (ref)
   (declare (type ref ref))
   (let ((leaf (ref-leaf ref)))
@@ -678,6 +685,57 @@
               (values 'array-in-bounds-p array-var index-constant)
               (values 'array-in-bounds-p index-var array-var)))))))
 
+(defun find-equality-constraint (operator x y not-p)
+  (let ((constraints (lambda-var-equality-constraints x)))
+    (when constraints
+      (loop for con across constraints
+            when (and (eq (equality-constraint-operator con) operator)
+                      (eq (constraint-not-p con) not-p)
+                      (eq (constraint-y con) y))
+            return con))))
+
+(defun find-or-create-equality-constraint (operator x y not-p)
+  (or (find-equality-constraint operator x y not-p)
+      (let ((new (make-equality-constraint (length *constraint-universe*)
+                                           operator
+                                           x y not-p)))
+        (vector-push-extend new *constraint-universe* (1+ (length *constraint-universe*)))
+        (flet ((add (var)
+                 (conset-adjoin new (lambda-var-constraints var))
+                 (macrolet ((ensure-vec (place)
+                              `(or ,place
+                                   (setf ,place
+                                         (make-array 8 :adjustable t :fill-pointer 0)))))
+                   (vector-push-extend new (ensure-vec (lambda-var-equality-constraints var)))
+                   (vector-push-extend new (ensure-vec (lambda-var-inheritable-constraints var))))))
+          (add x)
+          (when (lambda-var-p y)
+            (add y)))
+        new)))
+
+(defun add-equality-constraints (operator args constraints
+                                 consequent-constraints
+                                 alternative-constraints)
+  (case operator
+    ((eq eql)
+     (when (= (length args) 2)
+       (let* ((x (ok-lvar-lambda-var (first args) constraints))
+              (second (second args))
+              (y (if (constant-lvar-p second)
+                     (find-constant (lvar-value second))
+                     (ok-lvar-lambda-var second constraints))))
+         (when (and x y)
+           ;; TODO: inherit constraints
+           (let ((con (find-or-create-equality-constraint operator x y nil))
+                 (not-con (find-or-create-equality-constraint operator x y t)))
+             (conset-adjoin con consequent-constraints)
+             (conset-adjoin not-con alternative-constraints))))))))
+
+(defun add-eq-constraint (var lvar gen)
+  (let ((var2 (ok-lvar-lambda-var lvar gen)))
+    (when var2
+      (conset-adjoin (find-or-create-equality-constraint 'eq var var2 nil) gen))))
+
 ;;; Add test constraints to the consequent and alternative blocks of
 ;;; the test represented by USE.
 (defun add-test-constraints (use if constraints)
@@ -703,9 +761,12 @@
           (combination
            (unless (eq (combination-kind use)
                        :error)
-             (let ((name (lvar-fun-name
-                          (basic-combination-fun use)))
+             (let ((name (uncross
+                          (lvar-fun-name
+                           (basic-combination-fun use))))
                    (args (basic-combination-args use)))
+               (add-equality-constraints name args
+                                         constraints consequent-constraints alternative-constraints)
                (case name
                  ((%typep %instance-typep)
                   (let ((type (second args)))
@@ -771,6 +832,16 @@
                           (add name var1 (lvar-type arg2) nil))
                         (when var2
                           (add (if (eq name '<) '> '<) var2 (lvar-type arg1) nil))))))
+                 (=
+                  (when (= (length args) 2)
+                    (let* ((arg1 (first args))
+                           (var1 (ok-lvar-lambda-var arg1 constraints))
+                           (arg2 (second args))
+                           (var2 (ok-lvar-lambda-var arg2 constraints)))
+                      (when var1
+                        (add name var1 (lvar-type arg2) nil))
+                      (when var2
+                        (add name var2 (lvar-type arg1) nil)))))
                  (t
                   (add-combination-test-constraints use constraints
                                                     consequent-constraints
@@ -869,6 +940,22 @@
           (modified-numeric-type x :low new-bound)
           (modified-numeric-type x :high new-bound)))))
 
+(defun constrain-real-to-integer (y greater or-equal)
+  (declare (type numeric-type y))
+  (flet ((exclude (x)
+           (cond ((not x) nil)
+                 (or-equal x)
+                 (t (list x))))
+         (bound (x)
+           (if greater
+               (numeric-type-low x)
+               (numeric-type-high x))))
+    (let ((bound (exclude (bound y))))
+      (when bound
+        (if greater
+            (make-numeric-type :low bound)
+            (make-numeric-type :high bound))))))
+
 ;;; Return true if LEAF is "visible" from NODE.
 (defun leaf-visible-from-node-p (leaf node)
   (cond
@@ -921,6 +1008,10 @@
                (other (if (eq x leaf) y x))
                (kind (constraint-kind con)))
           (case kind
+            (equality
+             (unless (eq (ref-constraints ref)
+                         (pushnew con (ref-constraints ref)))
+               (reoptimize-lvar (node-lvar ref))))
             (typep
              (if not-p
                  (if (member-type-p other)
@@ -952,17 +1043,30 @@
                         (setq res (type-approx-intersection2
                                    res other-type))))))))
             ((< >)
-             (cond
-               ((and (integer-type-p res) (integer-type-p y))
-                (let ((greater (eq kind '>)))
-                  (let ((greater (if not-p (not greater) greater)))
-                    (setq res
-                          (constrain-integer-type res y greater not-p)))))
-               ((and (float-type-p res) (float-type-p y))
-                (let ((greater (eq kind '>)))
-                  (let ((greater (if not-p (not greater) greater)))
-                    (setq res
-                          (constrain-float-type res y greater not-p)))))))))))
+             (let* ((greater (eq kind '>))
+                    (greater (if not-p (not greater) greater)))
+               (cond
+                 ((and (integer-type-p res) (integer-type-p y))
+                  (setq res
+                        (constrain-integer-type res y greater not-p)))
+                 ((and (float-type-p res) (float-type-p y))
+                  (setq res
+                        (constrain-float-type res y greater not-p)))
+                 ((integer-type-p y)
+                  (let ((type (constrain-real-to-integer y greater not-p)))
+                    (when type
+                      (setf res
+                            (type-approx-intersection2 res type))))))))
+            (=
+             (when (and (numeric-type-p y)
+                        (not not-p))
+               (setf res
+                     (type-approx-intersection2 res
+                                                (type-union (make-numeric-type :low (numeric-type-low y)
+                                                                               :high (numeric-type-high y))
+                                                            (make-numeric-type :complexp :complex
+                                                                               :low (numeric-type-low y)
+                                                                               :high (numeric-type-high y)))))))))))
     (cond ((and (if-p (node-dest ref))
                 (or (xset-member-p nil not-set)
                     (csubtypep (specifier-type 'null) not-res)))
@@ -976,10 +1080,10 @@
              ;; type propagation for classes is not as important anyway.
              (cond #-sb-xc-host
                    ((and
-                     (eq sb!pcl::**boot-state** 'sb!pcl::complete)
+                     (eq sb-pcl::**boot-state** 'sb-pcl::complete)
                      (block nil
                        (let ((standard-object (find-classoid 'standard-object)))
-                         (sb!kernel::map-type
+                         (sb-kernel::map-type
                           (lambda (type)
                             (when (and (classoid-p type)
                                        (csubtypep type standard-object))
@@ -1069,7 +1173,8 @@
            (unless (eq type *universal-type*)
              (conset-add-constraint gen 'typep var type nil)))
          (unless (policy node (> compilation-speed speed))
-           (maybe-add-eql-var-var-constraint var (set-value node) gen))))
+           (maybe-add-eql-var-var-constraint var (set-value node) gen))
+         (add-eq-constraint var (set-value node) gen)))
       (combination
        (when (eq (combination-kind node) :known)
          (binding* ((info (combination-fun-info node) :exit-if-null)
