@@ -9,7 +9,7 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!DISASSEM")
+(in-package "SB-DISASSEM")
 
 ;;; types and defaults
 
@@ -34,7 +34,7 @@
 ;; and disassembler for other supported backends,
 ;; if that backend has been converted to use a distinct ASM package.
 (eval-when (:compile-toplevel :load-toplevel :execute)
-  (defparameter sb!assem::*backend-instruction-set-package*
+  (defparameter sb-assem::*backend-instruction-set-package*
     (find-package #.(sb-cold::backend-asm-package-name))))
 
 ;; How many columns of output to allow for the address preceding each line.
@@ -67,7 +67,7 @@
 ;;;; KLUDGE: It's not clear that using bit-vectors would be any more efficient.
 ;;;; Perhaps the abstraction could go away. -- WHN 19991124
 
-#!-sb-fluid
+#-sb-fluid
 (declaim (inline dchunk-or dchunk-and dchunk-clear dchunk-not
                  dchunk-make-mask dchunk-make-field
                  dchunk-extract
@@ -81,8 +81,9 @@
 ;;; Ideally this constant would go in the 'insts' file for the architecture,
 ;;; but there's really no easy way to do that at present.
 (defconstant dchunk-bits
-  #!+x86-64 56
-  #!-x86-64 sb!vm:n-word-bits)
+  #+x86-64 56
+  #+ppc64 32
+  #-(or x86-64 ppc64) sb-vm:n-word-bits)
 
 (deftype dchunk ()
   `(unsigned-byte ,dchunk-bits))
@@ -394,7 +395,7 @@
                                    instruction-format ~W bits wide.~:>"
                              (arg-name arg) bytespec format-length))
                     (correct-dchunk-bytespec-for-endianness
-                     bytespec format-length sb!c:*backend-byte-order*))
+                     bytespec format-length sb-c:*backend-byte-order*))
                   fields)))
   arg)
 
@@ -424,9 +425,17 @@
         (let* ((binding (cdr rendering))
                (vars (car binding))
                (vals (cdr binding)))
-          (if (listp vars)
-              (mapc (lambda (var val) (push `(,var ,val) bindings)) vars vals)
-              (push `(,vars ,vals) bindings)))))))
+          ;; We can end up here with VARS = NIL, and VALS = an atom.
+          ;; As the spec says, MAPC "should be prepared to signal an error
+          ;; ... if any list is not a proper list"
+          ;; We don't err in that situation because we check for ENDP of the
+          ;; lists from left to right. However, at least one implementation
+          ;; does rigorously use ENDP on both lists on each iteration.
+          (cond ((not vars))
+                ((listp vars)
+                 (mapc (lambda (var val) (push `(,var ,val) bindings)) vars vals))
+                (t
+                 (push `(,vars ,vals) bindings))))))))
 
 ;;; Return the form(s) that should be evaluated to render ARG in the chosen
 ;;; RENDERING style, which is one of :RAW, :SIGN-EXTENDED,
@@ -437,7 +446,7 @@
   (labels ((tempvars (n)
              (if (plusp n)
                  (cons (package-symbolicate
-                        (load-time-value (find-package "SB!DISASSEM"))
+                        #.(find-package "SB-DISASSEM")
                         ".T" (write-to-string (incf *!temp-var-counter*)))
                        (tempvars (1- n))))))
     (let* ((arg-cell (assq arg funstate))
@@ -502,7 +511,7 @@
                                unless (eq ind :type)
                                nconc (list ind val)))))
     `(progn
-       (eval-when (:compile-toplevel)
+       (eval-when (:compile-toplevel :execute)
          (%def-arg-type ',name ,inherit ,@(massage-arg pair :compile)))
        #-sb-xc-host ; Host doesn't need the real definition.
        (%def-arg-type ',name ,inherit ,@(massage-arg pair :eval)))))
@@ -563,20 +572,33 @@
         (chunk inst stream dstate
                &aux (chunk (truly-the dchunk chunk))
                     (inst (truly-the instruction inst))
-                    (stream (truly-the stream stream))
+                    (stream (truly-the (or null stream) stream))
                     (dstate (truly-the disassem-state dstate)))
        (macrolet ((local-format-arg (arg fmt)
                     `(funcall (formatter ,fmt) stream ,arg)))
+         ;; All the LOCAL-foo functions print nothing if stream is nil
          (flet ((local-tab-to-arg-column ()
                   (tab (dstate-argument-column dstate) stream))
                 (local-print-name ()
-                  (princ (inst-print-name inst) stream))
+                  (when stream (princ (inst-print-name inst) stream)))
+                (local-print-arg-separator ()
+                  (when stream (princ ", " stream)))
                 (local-write-char (ch)
-                  (write-char ch stream))
+                  (when stream (write-char ch stream)))
+                (local-princ-symbol (thing)
+                  ;; This special case is reqired so that decoding x86 JMP and CMOV
+                  ;; to an instruction model does not insert a superfluous symbol.
+                  ;; I hope this is right. We really only want to elide symbols
+                  ;; that are not inside a :COND or other selector.
+                  (when stream (princ thing stream)))
                 (local-princ (thing)
-                  (princ thing stream))
+                  (if stream
+                      (princ thing stream)
+                      (push thing (dstate-operands dstate))))
                 (local-princ16 (thing)
-                  (princ16 thing stream))
+                  (if stream
+                      (princ16 thing stream)
+                      (push thing (dstate-operands dstate))))
                 (local-call-arg-printer (arg printer)
                   (funcall printer arg stream dstate))
                 (local-call-global-printer (fun)
@@ -592,8 +614,8 @@
                 (adjust-label (val adjust-fun)
                   (funcall adjust-fun val dstate)))
            (declare (ignorable #'local-tab-to-arg-column
-                               #'local-print-name
-                               #'local-princ #'local-princ16
+                               #'local-print-name #'local-print-arg-separator
+                               #'local-princ #'local-princ16 #'local-princ-symbol
                                #'local-write-char
                                #'local-call-arg-printer
                                #'local-call-global-printer
@@ -601,7 +623,8 @@
                                #'local-filtered-value
                                #'lookup-label #'adjust-label)
                     (inline local-tab-to-arg-column
-                            local-princ local-princ16
+                            local-print-arg-separator
+                            local-princ local-princ16 local-princ-symbol
                             local-call-arg-printer local-call-global-printer
                             local-filtered-value local-extract
                             lookup-label adjust-label))
@@ -716,13 +739,6 @@
 ;;;; some simple functions that help avoid consing when we're just
 ;;;; recursively filtering things that usually don't change
 
-(defun recons (old-cons car cdr)
-  "If CAR is eq to the car of OLD-CONS and CDR is eq to the CDR, return
-  OLD-CONS, otherwise return (cons CAR CDR)."
-  (if (and (eq car (car old-cons)) (eq cdr (cdr old-cons)))
-      old-cons
-      (cons car cdr)))
-
 (defun sharing-mapcar (fun list)
   (declare (type function fun))
   "A simple (one list arg) mapcar that avoids consing up a new list
@@ -766,7 +782,15 @@
         ((symbolp source)
          (compile-print source funstate))
         ((atom source)
-         `(local-princ ',source))
+         ;; Skip the argument separator if disassembling into a model.
+         ;; The ", " string is quite consistent amongst backends.
+         ;; Making this any more general - such as never printing strings
+         ;; into a model - would be prone to information loss.
+         (if (and (stringp source) (string= source ", "))
+             `(local-print-arg-separator)
+             `(local-princ ',source)))
+        ((eq (car source) 'quote)
+         `(,(if (symbolp (cadr source)) 'local-princ-symbol 'local-princ) ,source))
         ((eq (car source) :using)
          (let ((f (cadr source)))
           (unless (typep f '(or string (cons (eql function) (cons symbol null))))
@@ -782,8 +806,6 @@
               (when (>= ,form 0)
                 (local-write-char #\+))
               (local-princ ,form))))
-        ((eq (car source) 'quote)
-         `(local-princ ,source))
         ((eq (car source) 'function)
          `(local-call-global-printer ,source))
         ((eq (car source) :cond)
@@ -851,6 +873,8 @@
           ((eq key :negative)
            `(< ,(arg-value-form (arg-or-lose subj funstate) funstate :numeric)
                0))
+          ((eq key :test)
+           `(,@body ,(arg-value-form (arg-or-lose subj funstate) funstate :numeric)))
           ((eq key :same-as)
            (let ((arg1 (arg-or-lose subj funstate))
                  (arg2 (arg-or-lose (car body) funstate)))
@@ -877,17 +901,16 @@
           (t
            (pd-error "bogus test-form: ~S" test)))))
 
-#!-sb-fluid (declaim (inline bytes-to-bits))
-(declaim (maybe-inline sign-extend tab tab0))
+#-sb-fluid (declaim (inline bytes-to-bits))
 
 (defun bytes-to-bits (bytes)
   (declare (type disassem-length bytes))
-  (* bytes sb!vm:n-byte-bits))
+  (* bytes sb-vm:n-byte-bits))
 
 (defun bits-to-bytes (bits)
   (declare (type disassem-length bits))
   (multiple-value-bind (bytes rbits)
-      (truncate bits sb!vm:n-byte-bits)
+      (truncate bits sb-vm:n-byte-bits)
     (when (not (zerop rbits))
       (error "~W bits is not a byte-multiple." bits))
     bytes))
@@ -898,13 +921,3 @@
   (if (logbitp (1- size) int)
       (dpb int (byte size 0) -1)
       int))
-
-(defun tab (column stream)
-  (funcall (formatter "~V,1t") stream column)
-  nil)
-(defun tab0 (column stream)
-  (funcall (formatter "~V,0t") stream column)
-  nil)
-
-(defun princ16 (value stream)
-  (write value :stream stream :radix t :base 16 :escape nil))
