@@ -16,7 +16,7 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!C")
+(in-package "SB-C")
 
 ;;;; storage class and storage base definition
 
@@ -43,7 +43,7 @@
            (error "A size specification is meaningless in a ~S SB." kind)))
         ((:finite :unbounded)
          (unless size (error "Size is not specified in a ~S SB." kind))
-         (aver (<= size sb!vm:finite-sc-offset-limit))
+         (aver (<= size sb-vm:finite-sc-offset-limit))
          (aver (= 1 (logcount size-alignment)))
          (aver (not (logtest size (1- size-alignment))))
          (aver (not (logtest size-increment (1- size-alignment))))))
@@ -199,7 +199,7 @@
 
      (defun ,name ,lambda-list
        (declare (ignorable ,(car lambda-list)))
-       (sb!assem:assemble ()
+       (sb-assem:assemble ()
          ,@body))))
 
 (defglobal *sc-vop-slots*
@@ -215,8 +215,7 @@
   (declare (type symbol name) (type list scs))
   (let ((scns (mapcar #'sc-number-or-lose scs)))
     `(progn
-       (/show0 "doing !DEF-PRIMITIVE-TYPE, NAME=..")
-       (/primitive-print ,(symbol-name name))
+       (/show "doing !DEF-PRIMITIVE-TYPE" ,(string name))
        (assert (not (gethash ',name *backend-primitive-type-names*)))
        (setf (gethash ',name *backend-primitive-type-names*)
              (make-primitive-type :name ',name
@@ -341,7 +340,8 @@
   ;; call/return VOPs
   (move-args nil :type (member nil :local-call :full-call :known-return))
   (args-var '.args. :type symbol)
-  (results-var '.results. :type symbol))
+  (results-var '.results. :type symbol)
+  (before-load :unspecified :type (or (member :unspecified) list)))
 (defprinter (vop-parse)
   name
   (inherits :test inherits)
@@ -368,7 +368,7 @@
 ;;; The list of slots in the structure, not including the OPERANDS slot.
 ;;; Order here is insignificant; it happens to be alphabetical.
 (defglobal vop-parse-slot-names
-    '(arg-types args args-var body conditional-p cost guard ignores info-args inherits
+    '(arg-types args args-var before-load body conditional-p cost guard ignores info-args inherits
       ltn-policy more-args more-results move-args name node-var note result-types
       results results-var save-p source-location temps translate variant variant-vars vop-var))
 ;; A sanity-check. Of course if this fails, the likelihood is that you can't even
@@ -409,6 +409,7 @@
                                        (vop-parse-or-lose name)))))))
 
        (let ((vop (template-or-lose ',name)))
+         (setf (vop-info-move-vop-p vop) t)
          (do-sc-pairs (from-sc to-sc ',scs)
            (dolist (dest-sc (cons to-sc (sc-alternate-scs to-sc)))
              (let ((vec (,accessor dest-sc)))
@@ -513,7 +514,7 @@
   (let ((temps (vop-parse-temps parse))
         (element-type '(unsigned-byte 16)))
     (when temps
-      (let ((results (!make-specialized-array (length temps) element-type))
+      (let ((results (sb-xc:make-array (length temps) :element-type element-type))
             (index 0))
         (dolist (temp temps)
           (declare (type operand-parse temp))
@@ -522,7 +523,7 @@
             (aver sc)
             (setf (aref results index)
                   (if offset
-                      (+ (ash offset (1+ sb!vm:sc-number-bits))
+                      (+ (ash offset (1+ sb-vm:sc-number-bits))
                          (ash (sc-number-or-lose sc) 1)
                          1)
                       (ash (sc-number-or-lose sc) 1))))
@@ -595,7 +596,7 @@
              ;; not correspond to the definition in
              ;; src/compiler/vop.lisp.
              (te-type '(unsigned-byte 16))
-             (ordering (!make-specialized-array (length sorted) oe-type)))
+             (ordering (sb-xc:make-array (length sorted) :element-type oe-type)))
         (let ((index 0))
           (dolist (ref sorted)
             (setf (aref ordering index) (cdr ref))
@@ -604,8 +605,7 @@
           :num-results ,num-results
           :ref-ordering ,ordering
           ,@(when (targets)
-              `(:targets ,(!make-specialized-array
-                           (length (targets)) te-type (targets)))))))))
+              `(:targets ,(sb-xc:coerce (targets) `(vector ,te-type)))))))))
 
 (defun make-emit-function-and-friends (parse)
   `(:temps ,(compute-temporaries-description parse)
@@ -725,7 +725,8 @@
         (n-args (vop-parse-args-var parse))
         (n-results (vop-parse-results-var parse))
         (operands (vop-parse-operands parse))
-        (n-info (gensym)) (n-variant (gensym)))
+        (n-info (gensym)) (n-variant (gensym))
+        (dummy (gensym)))
     (collect ((binds)
               (loads)
               (saves))
@@ -753,23 +754,28 @@
                 ,@(access-operands (vop-parse-args parse)
                                    (vop-parse-more-args parse)
                                    n-args)
-                  ,@(access-operands (vop-parse-results parse)
-                                     (vop-parse-more-results parse)
-                                     n-results)
-                  ,@(access-operands (vop-parse-temps parse) nil
-                                     `(vop-temps ,n-vop))
-                  ,@(when (vop-parse-info-args parse)
-                      `((,n-info (vop-codegen-info ,n-vop))
-                        ,@(mapcar (lambda (x) `(,x (pop ,n-info)))
-                                  (vop-parse-info-args parse))))
-                  ,@(when (vop-parse-variant-vars parse)
-                      `((,n-variant (vop-info-variant (vop-info ,n-vop)))
-                        ,@(mapcar (lambda (x) `(,x (pop ,n-variant)))
-                                  (vop-parse-variant-vars parse))))
-                  ,@(when (vop-parse-node-var parse)
-                      `((,(vop-parse-node-var parse) (vop-node ,n-vop))))
-                  ,@(binds))
-           (declare (ignore ,@(vop-parse-ignores parse))
+                ,@(access-operands (vop-parse-results parse)
+                                   (vop-parse-more-results parse)
+                                   n-results)
+                ,@(access-operands (vop-parse-temps parse) nil
+                                   `(vop-temps ,n-vop))
+                ,@(when (vop-parse-info-args parse)
+                    `((,n-info (vop-codegen-info ,n-vop))
+                      ,@(mapcar (lambda (x) `(,x (pop ,n-info)))
+                                (vop-parse-info-args parse))))
+                ,@(when (vop-parse-variant-vars parse)
+                    `((,n-variant (vop-info-variant (vop-info ,n-vop)))
+                      ,@(mapcar (lambda (x) `(,x (pop ,n-variant)))
+                                (vop-parse-variant-vars parse))))
+                ,@(when (vop-parse-node-var parse)
+                    `((,(vop-parse-node-var parse) (vop-node ,n-vop))))
+                ,@(and (neq (vop-parse-before-load parse) :unspecified)
+                       `((,dummy (progn
+                                   ,@(vop-parse-before-load parse)))))
+                ,@(binds))
+           (declare (ignore ,@(vop-parse-ignores parse)
+                            ,@(and (neq (vop-parse-before-load parse) :unspecified)
+                                   `(,dummy)))
                     (ignorable ,n-args ,n-results))
            ,@(loads)
            (assemble ()
@@ -780,11 +786,11 @@
 (defun make-operand-parse-temp ()
   (without-package-locks
    (intern (format nil "OPERAND-PARSE-TEMP-~D" *parse-vop-operand-count*)
-           (symbol-package '*parse-vop-operand-count*))))
+           #.(find-package "SB-C"))))
 (defun make-operand-parse-load-tn ()
   (without-package-locks
    (intern (format nil "OPERAND-PARSE-LOAD-TN-~D" *parse-vop-operand-count*)
-           (symbol-package '*parse-vop-operand-count*))))
+           #.(find-package "SB-C"))))
 
 ;;; Given a list of operand specifications as given to DEFINE-VOP,
 ;;; return a list of OPERAND-PARSE structures describing the fixed
@@ -975,11 +981,15 @@
         (:generator
             (setf (vop-parse-cost parse)
                   (vop-spec-arg spec 'unsigned-byte 1 nil))
-          (setf (vop-parse-body parse) (cddr spec)))
+         (setf (vop-parse-body parse) (cddr spec)))
+        (:before-load
+         (setf (vop-parse-before-load parse) (cdr spec)))
         (:info
          (setf (vop-parse-info-args parse) (rest spec)))
         (:ignore
-         (setf (vop-parse-ignores parse) (rest spec)))
+         (setf (vop-parse-ignores parse)
+               (append (vop-parse-ignores parse)
+                       (rest spec))))
         (:variant
          (setf (vop-parse-variant parse) (rest spec)))
         (:variant-vars
@@ -989,7 +999,7 @@
                  (make-list (length vars) :initial-element nil))))
         (:variant-cost
          (setf (vop-parse-cost parse) (vop-spec-arg spec 'unsigned-byte)))
-        ((:vop-var :args-ref-var :results-ref-var)
+        (:vop-var
          (setf (vop-parse-vop-var parse) (vop-spec-arg spec 'symbol)))
         (:args-var
          (setf (vop-parse-args-var parse) (vop-spec-arg spec 'symbol)))
@@ -1042,8 +1052,8 @@
 (defun compute-loading-costs (op load-p)
   (declare (type operand-parse op))
   (let ((scs (operand-parse-scs op))
-        (costs (make-array sb!vm:sc-number-limit :initial-element nil))
-        (load-scs (make-array sb!vm:sc-number-limit :initial-element nil)))
+        (costs (make-array sb-vm:sc-number-limit :initial-element nil))
+        (load-scs (make-array sb-vm:sc-number-limit :initial-element nil)))
     (dolist (sc-name (reverse scs))
       (let* ((load-sc (sc-or-lose sc-name))
              (load-scn (sc-number load-sc)))
@@ -1069,7 +1079,7 @@
               (unless (eq op-load t)
                 (pushnew load-scn (svref load-scs op-scn))))))
 
-        (dotimes (i sb!vm:sc-number-limit)
+        (dotimes (i sb-vm:sc-number-limit)
           (unless (svref costs i)
             (let ((op-sc (svref *backend-sc-numbers* i)))
               (when op-sc
@@ -1082,11 +1092,11 @@
     (values costs load-scs)))
 
 (defconstant-eqx +no-costs+
-    (make-array sb!vm:sc-number-limit :initial-element 0)
+    (make-array sb-vm:sc-number-limit :initial-element 0)
   #'equalp)
 
 (defconstant-eqx +no-loads+
-    (make-array sb!vm:sc-number-limit :initial-element t)
+    (make-array sb-vm:sc-number-limit :initial-element t)
   #'equalp)
 
 ;;; Pick off the case of operands with no restrictions.
@@ -1581,7 +1591,7 @@
               ((quotify-slots ()
                  (collect ((forms))
                    (dolist (x vop-parse-slot-names (cons 'list (forms)))
-                     (let ((reader (package-symbolicate (symbol-package 'vop-parse)
+                     (let ((reader (package-symbolicate (sb-xc:symbol-package 'vop-parse)
                                                         "VOP-PARSE-" x)))
                        (forms
                         (case x
